@@ -13,9 +13,6 @@ import pkg_resources
 import platform
 import re
 import signal
-import sys
-import traceback
-from types import TracebackType
 from typing import (
     Any,
     AsyncIterator,
@@ -34,13 +31,11 @@ from typing import (
     Sequence,
     Set,
     Tuple,
-    Type,
     TypeVar,
     Union,
     TYPE_CHECKING,
     cast,
 )
-import weakref
 
 import aioredis
 import aiotools
@@ -81,7 +76,6 @@ from ai.backend.common.types import (
 from ai.backend.common.events import (
     EventProducer,
     AbstractEvent,
-    AgentErrorEvent,
     AgentHeartbeatEvent,
     AgentStartedEvent,
     AgentTerminatedEvent,
@@ -99,7 +93,7 @@ from ai.backend.common.events import (
     SessionFailureEvent,
     SessionSuccessEvent,
 )
-from ai.backend.common.utils import cancel_tasks, current_loop
+from ai.backend.common.utils import current_loop
 from ai.backend.common.plugin.monitor import ErrorPluginContext, StatsPluginContext
 from ai.backend.common.service_ports import parse_service_ports
 from . import __version__ as VERSION
@@ -110,7 +104,6 @@ from .kernel import (
     KernelFeatures,
     match_distro_data,
 )
-from . import resources as resources_mod
 from .resources import (
     AbstractComputeDevice,
     AbstractComputePlugin,
@@ -196,20 +189,17 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
     zmq_ctx: zmq.asyncio.Context
 
     restarting_kernels: MutableMapping[KernelId, RestartTracker]
-    terminating_kernels: Set[KernelId]
     timer_tasks: MutableSequence[asyncio.Task]
-    container_lifecycle_queue: asyncio.Queue[ContainerLifecycleEvent | Sentinel]
+    container_lifecycle_queue: 'asyncio.Queue[Union[ContainerLifecycleEvent, Sentinel]]'
 
     stat_ctx: StatContext
     stat_sync_sockpath: Path
     stat_sync_task: asyncio.Task
 
-    stats_monitor: StatsPluginContext  # unused currently
-    error_monitor: ErrorPluginContext  # unused in favor of produce_error_event()
+    stats_monitor: StatsPluginContext
+    error_monitor: ErrorPluginContext
 
     _pending_creation_tasks: Dict[KernelId, Set[asyncio.Task]]
-    _ongoing_exec_batch_tasks: weakref.WeakSet[asyncio.Task]
-    _ongoing_destruction_tasks: weakref.WeakValueDictionary[KernelId, asyncio.Task]
 
     def __init__(
         self,
@@ -229,7 +219,6 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         self.computers = {}
         self.images = {}  # repoTag -> digest
         self.restarting_kernels = {}
-        self.terminating_kernels = set()
         self.stat_ctx = StatContext(
             self, mode=StatModes(local_config['container']['stats-type']),
         )
@@ -242,8 +231,6 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         self.error_monitor = error_monitor
         self._rx_distro = re.compile(r"\.([a-z-]+\d+\.\d+)\.")
         self._pending_creation_tasks = defaultdict(set)
-        self._ongoing_exec_batch_tasks = weakref.WeakSet()
-        self._ongoing_destruction_tasks = weakref.WeakValueDictionary()
 
     async def __ainit__(self) -> None:
         """
@@ -262,10 +249,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                 encoding=None,
             )
 
-        self.event_producer = await EventProducer.new(
-            connect_redis_event_db,
-            log_events=self.local_config['debug']['log-events'],
-        )
+        self.event_producer = await EventProducer.new(connect_redis_event_db)
         self.redis_stream_pool = await redis.connect_with_retries(
             self.local_config['redis']['addr'].as_sockaddr(),
             db=4,  # REDIS_STREAM_DB in backend.ai-manager
@@ -284,7 +268,6 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         ipc_base_path.mkdir(parents=True, exist_ok=True)
         self.zmq_ctx = zmq.asyncio.Context()
 
-        resources_mod.log_alloc_map = self.local_config['debug']['log-alloc-map']
         computers, self.slots = await self.detect_resources()
         for name, computer in computers.items():
             devices = await computer.list_devices()
@@ -317,8 +300,6 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         An implementation of AbstractAgent would define its own ``shutdown()`` method.
         It must call this super method in an appropriate order, only once.
         """
-        await cancel_tasks(self._ongoing_exec_batch_tasks)
-
         # Close all pending kernel runners.
         for kernel_obj in self.kernel_registry.values():
             if kernel_obj.runner is not None:
@@ -328,8 +309,10 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
             await self.clean_all_kernels(blocking=True)
 
         # Stop timers.
-        cancel_results = await cancel_tasks(self.timer_tasks)
-        for result in cancel_results:
+        for task in self.timer_tasks:
+            task.cancel()
+        timer_cancel_results = await asyncio.gather(*self.timer_tasks, return_exceptions=True)
+        for result in timer_cancel_results:
             if isinstance(result, Exception):
                 log.error('timer cancellation error: {}', result)
 
@@ -357,62 +340,50 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
             _log = log.debug if isinstance(event, AgentHeartbeatEvent) else log.info
         else:
             _log = (lambda *args: None) if isinstance(event, AgentHeartbeatEvent) else log.info
-        if self.local_config['debug']['log-events']:
-            _log('produce_event({0})', event)
+        _log('produce_event({0})', event)
         if isinstance(event, KernelTerminatedEvent):
             pending_creation_tasks = self._pending_creation_tasks.get(event.kernel_id, None)
             if pending_creation_tasks is not None:
                 for t in pending_creation_tasks:
-                    if not t.done():
-                        t.cancel()
-                        await t
+                    t.cancel()
         await self.event_producer.produce_event(event, source=self.local_config['agent']['id'])
 
-    async def produce_error_event(
-        self,
-        exc_info: Tuple[Type[BaseException], BaseException, TracebackType] = None,
-    ) -> None:
-        exc_type, exc, tb = sys.exc_info() if exc_info is None else exc_info
-        pretty_message = ''.join(traceback.format_exception_only(exc_type, exc)).strip()
-        pretty_tb = ''.join(traceback.format_tb(tb)).strip()
-        await self.produce_event(AgentErrorEvent(pretty_message, pretty_tb))
-
     async def heartbeat(self, interval: float):
         """
         Send my status information and available kernel images to the manager(s).
         """
         res_slots = {}
+        for cctx in self.computers.values():
+            for slot_key, slot_type in cctx.instance.slot_types:
+                res_slots[slot_key] = (
+                    slot_type,
+                    str(self.slots.get(slot_key, 0)),
+                )
+        agent_info = {
+            'ip': str(self.local_config['agent']['rpc-listen-addr'].host),
+            'region': self.local_config['agent']['region'],
+            'scaling_group': self.local_config['agent']['scaling-group'],
+            'addr': f"tcp://{self.local_config['agent']['rpc-listen-addr']}",
+            'resource_slots': res_slots,
+            'version': VERSION,
+            'compute_plugins': {
+                key: {
+                    'version': computer.instance.get_version(),
+                    **(await computer.instance.extra_info())
+                }
+                for key, computer in self.computers.items()
+            },
+            'images': snappy.compress(msgpack.packb([
+                (repo_tag, digest) for repo_tag, digest in self.images.items()
+            ])),
+        }
         try:
-            for cctx in self.computers.values():
-                for slot_key, slot_type in cctx.instance.slot_types:
-                    res_slots[slot_key] = (
-                        slot_type,
-                        str(self.slots.get(slot_key, 0)),
-                    )
-            agent_info = {
-                'ip': str(self.local_config['agent']['rpc-listen-addr'].host),
-                'region': self.local_config['agent']['region'],
-                'scaling_group': self.local_config['agent']['scaling-group'],
-                'addr': f"tcp://{self.local_config['agent']['rpc-listen-addr']}",
-                'resource_slots': res_slots,
-                'version': VERSION,
-                'compute_plugins': {
-                    key: {
-                        'version': computer.instance.get_version(),
-                        **(await computer.instance.extra_info())
-                    }
-                    for key, computer in self.computers.items()
-                },
-                'images': snappy.compress(msgpack.packb([
-                    (repo_tag, digest) for repo_tag, digest in self.images.items()
-                ])),
-            }
             await self.produce_event(AgentHeartbeatEvent(agent_info))
         except asyncio.TimeoutError:
             log.warning('event dispatch timeout: instance_heartbeat')
         except Exception:
             log.exception('instance_heartbeat failure')
-            await self.produce_error_event()
+            await self.error_monitor.capture_exception()
 
     async def collect_logs(
         self,
@@ -426,26 +397,25 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         chunk_buffer = BytesIO()
         chunk_length = 0
         try:
-            async with aiotools.aclosing(async_log_iterator):
-                async for fragment in async_log_iterator:
-                    fragment_length = len(fragment)
-                    chunk_buffer.write(fragment)
-                    chunk_length += fragment_length
-                    log_length += fragment_length
-                    while chunk_length >= chunk_size:
-                        cb = chunk_buffer.getbuffer()
-                        stored_chunk = bytes(cb[:chunk_size])
-                        await redis.execute_with_retries(
-                            lambda: self.redis_stream_pool.rpush(
-                                log_key, stored_chunk)
-                        )
-                        remaining = cb[chunk_size:]
-                        chunk_length = len(remaining)
-                        next_chunk_buffer = BytesIO(remaining)
-                        next_chunk_buffer.seek(0, SEEK_END)
-                        del remaining, cb
-                        chunk_buffer.close()
-                        chunk_buffer = next_chunk_buffer
+            async for fragment in async_log_iterator:
+                fragment_length = len(fragment)
+                chunk_buffer.write(fragment)
+                chunk_length += fragment_length
+                log_length += fragment_length
+                while chunk_length >= chunk_size:
+                    cb = chunk_buffer.getbuffer()
+                    stored_chunk = bytes(cb[:chunk_size])
+                    await redis.execute_with_retries(
+                        lambda: self.redis_stream_pool.rpush(
+                            log_key, stored_chunk)
+                    )
+                    remaining = cb[chunk_size:]
+                    chunk_length = len(remaining)
+                    next_chunk_buffer = BytesIO(remaining)
+                    next_chunk_buffer.seek(0, SEEK_END)
+                    del remaining, cb
+                    chunk_buffer.close()
+                    chunk_buffer = next_chunk_buffer
             assert chunk_length < chunk_size
             if chunk_length > 0:
                 await redis.execute_with_retries(
@@ -472,7 +442,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
             pass
         except Exception:
             log.exception('unhandled exception while syncing node stats')
-            await self.produce_error_event()
+            await self.error_monitor.capture_exception()
 
     async def collect_container_stat(self, interval: float):
         if self.local_config['debug']['log-stats']:
@@ -493,7 +463,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
             pass
         except Exception:
             log.exception('unhandled exception while syncing container stats')
-            await self.produce_error_event()
+            await self.error_monitor.capture_exception()
 
     async def _handle_start_event(self, ev: ContainerLifecycleEvent) -> None:
         kernel_obj = self.kernel_registry.get(ev.kernel_id)
@@ -503,11 +473,6 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
     async def _handle_destroy_event(self, ev: ContainerLifecycleEvent) -> None:
         result = None
         try:
-            current_task = asyncio.current_task()
-            assert current_task is not None
-            if ev.kernel_id not in self._ongoing_destruction_tasks:
-                self._ongoing_destruction_tasks[ev.kernel_id] = current_task
-            self.terminating_kernels.add(ev.kernel_id)
             kernel_obj = self.kernel_registry.get(ev.kernel_id)
             if kernel_obj is None:
                 log.warning('destroy_kernel(k:{0}) kernel missing (already dead?)',
@@ -520,7 +485,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                         )
                     return
                 else:
-                    await self.container_lifecycle_queue.put(
+                    self.container_lifecycle_queue.put_nowait(
                         ContainerLifecycleEvent(
                             ev.kernel_id,
                             ev.container_id,
@@ -535,22 +500,15 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                 if kernel_obj.runner is not None:
                     await kernel_obj.runner.close()
             result = await self.destroy_kernel(ev.kernel_id, ev.container_id)
-        except asyncio.CancelledError:
-            pass
         except Exception:
             log.exception('unhandled exception while processing DESTROY event')
-            await self.produce_error_event()
+            await self.error_monitor.capture_exception()
         finally:
             if ev.done_event is not None:
                 ev.done_event.set()
                 setattr(ev.done_event, '_result', result)
 
     async def _handle_clean_event(self, ev: ContainerLifecycleEvent) -> None:
-        destruction_task = self._ongoing_destruction_tasks.get(ev.kernel_id, None)
-        if destruction_task is not None and not destruction_task.done():
-            # let the destruction task finish first
-            await destruction_task
-            del destruction_task
         result = None
         try:
             kernel_obj = self.kernel_registry.get(ev.kernel_id)
@@ -562,7 +520,8 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                 ev.kernel_id in self.restarting_kernels,
             )
         except Exception:
-            await self.produce_error_event()
+            log.exception('unhandled exception while processing CLEAN event')
+            await self.error_monitor.capture_exception()
         finally:
             try:
                 kernel_obj = self.kernel_registry.get(ev.kernel_id)
@@ -595,7 +554,6 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                         await self.produce_event(
                             KernelTerminatedEvent(ev.kernel_id, ev.reason)
                         )
-                self.terminating_kernels.discard(ev.kernel_id)
 
     async def process_lifecycle_events(self) -> None:
         async with aiotools.TaskGroup() as tg:
@@ -607,8 +565,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                     return
                 # attr currently does not support customizing getstate/setstate dunder methods
                 # until the next release.
-                if self.local_config['debug']['log-events']:
-                    log.info(f'lifecycle event: {ev!r}')
+                log.info(f'lifecycle event: {ev!r}')
                 try:
                     if ev.event == LifecycleEvent.START:
                         tg.create_task(self._handle_start_event(ev))
@@ -619,9 +576,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                     else:
                         log.warning('unsupported lifecycle event: {!r}', ev)
                 except Exception:
-                    log.exception(
-                        'unexpected error in process_lifecycle_events(): {!r}, continuing...', ev,
-                    )
+                    log.exception('unexpected error in process_lifecycle_events(), continuing...')
                 finally:
                     self.container_lifecycle_queue.task_done()
 
@@ -652,7 +607,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
             container_id = kernel_obj['container_id']
         except KeyError:
             pass
-        await self.container_lifecycle_queue.put(
+        self.container_lifecycle_queue.put_nowait(
             ContainerLifecycleEvent(
                 kernel_id,
                 container_id,
@@ -679,18 +634,10 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                 computer_set.alloc_map.clear()
             for kernel_id, container in (await self.enumerate_containers()):
                 for computer_set in self.computers.values():
-                    try:
-                        await computer_set.instance.restore_from_container(
-                            container,
-                            computer_set.alloc_map,
-                        )
-                    except Exception:
-                        log.warning(
-                            "rescan_resoucre_usage(k:{}): "
-                            "failed to read kernel resource info; "
-                            "maybe already terminated",
-                            kernel_id,
-                        )
+                    await computer_set.instance.restore_from_container(
+                        container,
+                        computer_set.alloc_map,
+                    )
 
     async def sync_container_lifecycles(self, interval: float) -> None:
         """
@@ -706,9 +653,9 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
             try:
                 # Check if: there are dead containers
                 for kernel_id, container in (await self.enumerate_containers(DEAD_STATUS_SET)):
-                    if kernel_id in self.restarting_kernels or kernel_id in self.terminating_kernels:
+                    if kernel_id in self.restarting_kernels:
                         continue
-                    log.info('detected dead container during lifeycle sync (k:{}, c:{})',
+                    log.info('detected dead container during lifeycle sync (k:{0}, c:{})',
                             kernel_id, container.id)
                     terminated_kernels[kernel_id] = ContainerLifecycleEvent(
                         kernel_id,
@@ -722,7 +669,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                     known_kernels[kernel_id] = kernel_obj['container_id']
                 # Check if: kernel_registry has the container but it's gone.
                 for kernel_id in (known_kernels.keys() - alive_kernels.keys()):
-                    if kernel_id in self.restarting_kernels or kernel_id in self.terminating_kernels:
+                    if kernel_id in self.restarting_kernels:
                         continue
                     terminated_kernels[kernel_id] = ContainerLifecycleEvent(
                         kernel_id,
@@ -743,7 +690,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
             finally:
                 # Enqueue the events.
                 for kernel_id, ev in terminated_kernels.items():
-                    await self.container_lifecycle_queue.put(ev)
+                    self.container_lifecycle_queue.put_nowait(ev)
 
     async def clean_all_kernels(self, blocking: bool = False) -> None:
         kernel_ids = [*self.kernel_registry.keys()]
@@ -868,7 +815,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                         container_id=container.id,
                     )
                 elif container.status in DEAD_STATUS_SET:
-                    log.info('detected dead container while agent is down (k:{}, c:{})',
+                    log.info('detected dead container while agent is down (k:{0}, c:{})',
                              kernel_id, container.id)
                     await self.inject_container_lifecycle_event(
                         kernel_id,
@@ -1172,9 +1119,6 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         dotfile_extractor_path = Path(pkg_resources.resource_filename(
             'ai.backend.runner', 'extract_dotfiles.py'
         ))
-        persistent_files_warning_doc_path = Path(pkg_resources.resource_filename(
-            'ai.backend.runner', 'DO_NOT_STORE_PERSISTENT_FILES_HERE.md'
-        ))
 
         if matched_libc_style == 'musl':
             terminfo_path = Path(pkg_resources.resource_filename(
@@ -1194,8 +1138,6 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         _mount(MountTypes.BIND, tmux_path.resolve(), '/opt/kernel/tmux')
         _mount(MountTypes.BIND, sftp_server_path.resolve(), '/usr/libexec/sftp-server')
         _mount(MountTypes.BIND, scp_path.resolve(), '/usr/bin/scp')
-        _mount(MountTypes.BIND, persistent_files_warning_doc_path.resolve(),
-               '/home/work/DO_NOT_STORE_PERSISTENT_FILES_HERE.md')
 
         _mount(MountTypes.VOLUME, krunner_volume, '/opt/backend.ai')
         pylib_path = f'/opt/backend.ai/lib/python{krunner_pyver}/site-packages/'
@@ -1304,8 +1246,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
             )
 
         # Initialize the creation context
-        if self.local_config['debug']['log-kernel-config']:
-            log.debug('Kernel creation config: {0}', pretty(kernel_config))
+        log.debug('Kernel creation config: {0}', pretty(kernel_config))
         ctx = await self.create_kernel__init_context(
             kernel_id, kernel_config,
             restarting=restarting,
@@ -1322,6 +1263,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
             await self.create_kernel__get_extra_envs(ctx)
         )
         image_labels = kernel_config['image']['labels']
+        log.debug('image labels:\n{}', pretty(image_labels))
 
         # Check if we need to pull the container image
         do_pull = await self.check_image(
@@ -1398,7 +1340,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
         label_envs_corecount = image_labels.get('ai.backend.envs.corecount', '')
         envs_corecount = label_envs_corecount.split(',') if label_envs_corecount else []
         cpu_core_count = len(resource_spec.allocations[DeviceName('cpu')][SlotName('cpu')])
-        environ.update({k: str(cpu_core_count) for k in envs_corecount if k not in environ})
+        environ.update({k: str(cpu_core_count) for k in envs_corecount})
 
         # Realize mounts.
         await self.create_kernel__process_mounts(
@@ -1481,9 +1423,8 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                 json.dumps(cluster_info).encode('utf8'),
             )
 
-        if self.local_config['debug']['log-kernel-config']:
-            log.info('kernel starting with resource spec: \n{0}',
-                     pretty(attr.asdict(resource_spec)))
+        log.info('kernel starting with resource spec: \n{0}',
+                 pretty(attr.asdict(resource_spec)))
         kernel_obj = await self.create_kernel__spawn(
             ctx,
             resource_spec,
@@ -1494,6 +1435,10 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
             cmdargs,
         )
         self.kernel_registry[ctx.kernel_id] = kernel_obj
+        log.debug('kernel repl-in address: {0}:{1}',
+                  kernel_obj['kernel_host'], kernel_obj['repl_in_port'])
+        log.debug('kernel repl-out address: {0}:{1}',
+                  kernel_obj['kernel_host'], kernel_obj['repl_out_port'])
 
         current_task = asyncio.current_task()
         assert current_task is not None
@@ -1513,8 +1458,7 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
                         if live_service['name'] == service_port['name']:
                             service_port.update(live_service)
                             break
-            if self.local_config['debug']['log-kernel-config']:
-                log.debug('service ports:\n{!r}', pretty(service_ports))
+            log.debug('service ports:\n{!r}', pretty(service_ports))
         except (asyncio.CancelledError, zmq.error.ZMQError):
             log.warning("cancelled waiting of container startup (k:{})", kernel_id)
             raise RuntimeError("cancelled waiting of container startup due to "
@@ -1529,13 +1473,6 @@ class AbstractAgent(aobject, Generic[KernelObjectType, KernelCreationContextType
             KernelStartedEvent(kernel_id, creation_id)
         )
 
-        if kernel_config['session_type'] == 'batch' and kernel_config['cluster_role'] == 'main':
-            self._ongoing_exec_batch_tasks.add(
-                asyncio.create_task(
-                    self.execute_batch(kernel_id, kernel_config['startup_command'] or "")
-                )
-            )
-
         # The startup command for the batch-type sessions will be executed by the manager
         # upon firing of the "session_started" event.
 
