@@ -14,7 +14,6 @@ import secrets
 import shutil
 import signal
 import struct
-from subprocess import CalledProcessError
 import sys
 from typing import (
     Any,
@@ -254,7 +253,6 @@ class DockerAgent(AbstractAgent[DockerKernel, DockerKernelCreationContext]):
         for container in (await self.docker.containers.list()):
 
             async def _fetch_container_info(container):
-                kernel_id = "(unknown)"
                 try:
                     kernel_id = await get_kernel_id_from_container(container)
                     if kernel_id is None:
@@ -268,12 +266,11 @@ class DockerAgent(AbstractAgent[DockerKernel, DockerKernelCreationContext]):
                             )
                         )
                 except asyncio.CancelledError:
-                    pass
+                    raise
                 except Exception:
                     log.exception(
-                        "error while fetching container information (cid:{}, k:{})",
-                        container._id, kernel_id,
-                    )
+                        'error while fetching container information (cid: {})',
+                        container._id)
 
             fetch_tasks.append(_fetch_container_info(container))
 
@@ -676,8 +673,8 @@ class DockerAgent(AbstractAgent[DockerKernel, DockerKernelCreationContext]):
                     uid = self.local_config['container']['kernel-uid']
                     gid = self.local_config['container']['kernel-gid']
                     if os.geteuid() == 0:  # only possible when I am root.
-                        os.chown(str(priv_key_path), uid, gid)
-                        os.chown(str(pub_key_path), uid, gid)
+                        priv_key_path.chown(uid, gid)
+                        pub_key_path.chown(uid, gid)
                 priv_key_path.chmod(0o600)
             except Exception:
                 log.exception('error while writing cluster keypair')
@@ -870,8 +867,7 @@ class DockerAgent(AbstractAgent[DockerKernel, DockerKernelCreationContext]):
             'EntryPoint': ["/opt/kernel/entrypoint.sh"],
             'Cmd': cmdargs,
             'Env': [f'{k}={v}' for k, v in environ.items()],
-            'WorkingDir': "/home/work",
-            'Hostname': ctx.kernel_config['cluster_hostname'],
+            'WorkingDir': '/home/work',
             'Labels': {
                 'ai.backend.kernel-id': str(ctx.kernel_id),
                 'ai.backend.internal.block-service-ports':
@@ -918,8 +914,7 @@ class DockerAgent(AbstractAgent[DockerKernel, DockerKernelCreationContext]):
                 image_labels['ai.backend.service-ports'] + ',' + encoded_preopen_ports
         update_nested_dict(container_config, ctx.computer_docker_args)
         kernel_name = f"kernel.{ctx.image_ref.name.split('/')[-1]}.{ctx.kernel_id}"
-        if self.local_config['debug']['log-kernel-config']:
-            log.debug('full container config: {!r}', pretty(container_config))
+        log.debug('full container config: {!r}', pretty(container_config))
 
         # We are all set! Create and start the container.
         try:
@@ -1065,33 +1060,29 @@ class DockerAgent(AbstractAgent[DockerKernel, DockerKernelCreationContext]):
             container = self.docker.containers.container(container_id)
 
             async def log_iter():
-                it = container.log(
+                async for line in container.log(
                     stdout=True, stderr=True, follow=True,
-                )
-                async with aiotools.aclosing(it):
-                    async for line in it:
-                        yield line.encode('utf-8')
+                ):
+                    yield line.encode('utf-8')
 
             try:
                 with timeout(60):
                     await self.collect_logs(kernel_id, container_id, log_iter())
             except asyncio.TimeoutError:
-                log.warning('timeout for collecting container logs (k:{}, cid:{})',
-                            kernel_id, container_id)
+                log.warning('timeout for collecting container logs (cid:{})', container_id)
             except Exception as e:
-                log.warning('error while collecting container logs (k:{}, cid:{})',
-                            kernel_id, container_id, exc_info=e)
+                log.warning('error while collecting container logs (cid:{})',
+                            container_id, exc_info=e)
 
         kernel_obj = self.kernel_registry.get(kernel_id)
         if kernel_obj is not None:
             for domain_socket_proxy in kernel_obj.get('domain_socket_proxies', []):
-                if domain_socket_proxy.proxy_server.is_serving():
-                    domain_socket_proxy.proxy_server.close()
-                    await domain_socket_proxy.proxy_server.wait_closed()
-                    try:
-                        domain_socket_proxy.host_proxy_path.unlink()
-                    except IOError:
-                        pass
+                domain_socket_proxy.proxy_server.close()
+                await domain_socket_proxy.proxy_server.wait_closed()
+                try:
+                    domain_socket_proxy.host_proxy_path.unlink()
+                except IOError:
+                    pass
 
         if not self.local_config['debug']['skip-container-deletion'] and container_id is not None:
             container = self.docker.containers.container(container_id)
@@ -1122,8 +1113,6 @@ class DockerAgent(AbstractAgent[DockerKernel, DockerKernelCreationContext]):
                     await destroy_scratch_filesystem(tmp_dir)
                     await loop.run_in_executor(None, shutil.rmtree, tmp_dir)
                 await loop.run_in_executor(None, shutil.rmtree, scratch_dir)
-            except CalledProcessError:
-                pass
             except FileNotFoundError:
                 pass
 
@@ -1198,16 +1187,13 @@ class DockerAgent(AbstractAgent[DockerKernel, DockerKernelCreationContext]):
                         if evdata['Type'] != 'container':
                             # Our interest is the container-related events
                             continue
+                        if self.local_config['debug']['log-docker-events']:
+                            log.debug('docker-event: action={}, actor={}',
+                                      evdata['Action'], evdata['Actor'])
                         container_name = evdata['Actor']['Attributes']['name']
                         kernel_id = await get_kernel_id_from_container(container_name)
                         if kernel_id is None:
                             continue
-                        if (
-                            self.local_config['debug']['log-docker-events']
-                            and evdata['Action'] in ('start', 'die')
-                        ):
-                            log.debug('docker-event: action={}, actor={}',
-                                      evdata['Action'], evdata['Actor'])
                         if evdata['Action'] == 'start':
                             await asyncio.shield(handle_action_start(kernel_id, evdata))
                         elif evdata['Action'] == 'die':
