@@ -18,7 +18,6 @@ from typing import (
 import aiohttp
 from aiodocker.docker import Docker, DockerContainer
 from aiodocker.exceptions import DockerError
-import async_timeout
 import psutil
 
 from ai.backend.common.logging import BraceStyleAdapter
@@ -117,7 +116,7 @@ class CPUPlugin(AbstractComputePlugin):
     async def list_devices(self) -> Collection[CPUDevice]:
         cores = await libnuma.get_available_cores()
         overcommit_factor = int(os.environ.get('BACKEND_CPU_OVERCOMMIT_FACTOR', '1'))
-        assert 1 <= overcommit_factor <= 10
+        assert 1 <= overcommit_factor <= 4
         return [
             CPUDevice(
                 device_id=DeviceId(str(core_idx)),
@@ -189,11 +188,7 @@ class CPUPlugin(AbstractComputePlugin):
 
         async def api_impl(container_id):
             container = DockerContainer(ctx.agent.docker, id=container_id)
-            try:
-                async with async_timeout.timeout(2.0):
-                    ret = await fetch_api_stats(container)
-            except asyncio.TimeoutError:
-                return None
+            ret = await fetch_api_stats(container)
             if ret is None:
                 return None
             cpu_used = nmget(ret, 'cpu_stats.cpu_usage.total_usage', 0) / 1e6
@@ -463,11 +458,7 @@ class MemoryPlugin(AbstractComputePlugin):
 
         async def api_impl(container_id):
             container = DockerContainer(ctx.agent.docker, id=container_id)
-            try:
-                async with async_timeout.timeout(2.0):
-                    ret = await fetch_api_stats(container)
-            except asyncio.TimeoutError:
-                return None
+            ret = await fetch_api_stats(container)
             if ret is None:
                 return None
             mem_cur_bytes = nmget(ret, 'memory_stats.usage', 0)
