@@ -10,6 +10,7 @@ from pathlib import Path
 from pprint import pformat, pprint
 import signal
 import sys
+import traceback
 from typing import (
     Any,
     AsyncGenerator,
@@ -26,7 +27,6 @@ from typing import (
 )
 from uuid import UUID
 
-import aiomonitor
 import aiotools
 from aiotools import aclosing
 from callosum.rpc import Peer, RPCMessage
@@ -38,6 +38,7 @@ from trafaret.dataerror import DataError as TrafaretDataError
 
 from ai.backend.common import config, utils, identity, msgpack
 from ai.backend.common.etcd import AsyncEtcd, ConfigScopes
+from ai.backend.common.events import AgentErrorEvent
 from ai.backend.common.logging import Logger, BraceStyleAdapter
 from ai.backend.common.plugin.monitor import ErrorPluginContext, StatsPluginContext
 from ai.backend.common.types import (
@@ -111,7 +112,10 @@ def collect_error(meth: Callable) -> Callable:
         try:
             return await meth(self, *args, **kwargs)
         except Exception:
-            await self.agent.produce_error_event()
+            exc_type, exc, tb = sys.exc_info()
+            pretty_message = ''.join(traceback.format_exception_only(exc_type, exc)).strip()
+            pretty_tb = ''.join(traceback.format_tb(tb)).strip()
+            await self.agent.produce_event(AgentErrorEvent(pretty_message, pretty_tb))
             raise
     return _inner
 
@@ -179,8 +183,6 @@ class AgentRPCServer(aobject):
 
     async def __ainit__(self) -> None:
         # Start serving requests.
-        self._create_sema = asyncio.Semaphore(4)
-        self._destroy_sema = asyncio.Semaphore(4)
         await self.update_status('starting')
 
         if not self.skip_detect_manager:
@@ -315,30 +317,27 @@ class AgentRPCServer(aobject):
         raw_configs: Sequence[dict],
         raw_cluster_info: dict,
     ):
-        async with self._create_sema:
-            cluster_info = cast(ClusterInfo, raw_cluster_info)
-            session_id = SessionId(UUID(raw_session_id))
-            raw_results = []
-            coros = []
-            for raw_kernel_id, raw_config in zip(raw_kernel_ids, raw_configs):
-                log.info('rpc::create_kernel(k:{0}, img:{1})',
-                        raw_kernel_id, raw_config['image']['canonical'])
-                kernel_id = KernelId(UUID(raw_kernel_id))
-                kernel_config = cast(KernelCreationConfig, raw_config)
-                coros.append(self.agent.create_kernel(
-                    creation_id,
-                    session_id,
-                    kernel_id,
-                    kernel_config,
-                    cluster_info,
-                ))
-            results = await asyncio.gather(*coros, return_exceptions=True)
+        cluster_info = cast(ClusterInfo, raw_cluster_info)
+        session_id = SessionId(UUID(raw_session_id))
+        raw_results = []
+        coros = []
+        for raw_kernel_id, raw_config in zip(raw_kernel_ids, raw_configs):
+            log.info('rpc::create_kernel(k:{0}, img:{1})',
+                     raw_kernel_id, raw_config['image']['canonical'])
+            kernel_id = KernelId(UUID(raw_kernel_id))
+            kernel_config = cast(KernelCreationConfig, raw_config)
+            coros.append(self.agent.create_kernel(
+                creation_id,
+                session_id,
+                kernel_id,
+                kernel_config,
+                cluster_info,
+            ))
+        results = await asyncio.gather(*coros, return_exceptions=True)
         errors = [*filter(lambda item: isinstance(item, Exception), results)]
         if errors:
             # Raise up the first error.
-            if len(errors) == 1:
-                raise errors[0]
-            raise aiotools.MultiError("agent.create_kernels() failed", errors=errors)
+            raise errors[0]
         raw_results = [
             {
                 'id': str(result['id']),
@@ -364,17 +363,16 @@ class AgentRPCServer(aobject):
         reason: str = None,
         suppress_events: bool = False,
     ):
-        async with self._destroy_sema:
-            log.info('rpc::destroy_kernel(k:{0})', kernel_id)
-            done = asyncio.Event()
-            await self.agent.inject_container_lifecycle_event(
-                KernelId(UUID(kernel_id)),
-                LifecycleEvent.DESTROY,
-                reason or 'user-requested',
-                done_event=done,
-                suppress_events=suppress_events,
-            )
-            await done.wait()
+        log.info('rpc::destroy_kernel(k:{0})', kernel_id)
+        done = asyncio.Event()
+        await self.agent.inject_container_lifecycle_event(
+            KernelId(UUID(kernel_id)),
+            LifecycleEvent.DESTROY,
+            reason or 'user-requested',
+            done_event=done,
+            suppress_events=suppress_events,
+        )
+        await done.wait()
         return getattr(done, '_result', None)
 
     @rpc_function
@@ -448,7 +446,7 @@ class AgentRPCServer(aobject):
         kernel_id,          # type: str
         startup_command,    # type: str
     ) -> None:
-        # DEPRECATED
+        # TODO: use task-group to keep track of completion/cancellation
         asyncio.create_task(self.agent.execute_batch(
             KernelId(UUID(kernel_id)),
             startup_command,
@@ -622,16 +620,6 @@ async def server_main(
     # Pre-load compute plugin configurations.
     local_config['plugins'] = await etcd.get_prefix_dict('config/plugins/accelerator')
 
-    # Start aiomonitor.
-    # Port is set by config (default=50002).
-    monitor = aiomonitor.Monitor(
-        loop,
-        port=local_config['agent']['aiomonitor-port'],
-        console_enabled=False
-    )
-    monitor.prompt = "monitor (agent) >>> "
-    monitor.start()
-
     # Start RPC server.
     global agent_instance
     agent = await AgentRPCServer.new(
@@ -641,12 +629,9 @@ async def server_main(
     agent_instance = agent
 
     # Run!
-    try:
-        async with agent:
-            stop_signal = yield
-            agent.mark_stop_signal(stop_signal)
-    finally:
-        monitor.close()
+    async with agent:
+        stop_signal = yield
+        agent.mark_stop_signal(stop_signal)
 
 
 @click.group(invoke_without_command=True)
