@@ -14,6 +14,12 @@ from ai.backend.common.types import (
 )
 
 
+class AgentBackend(enum.Enum):
+    # The list of importable backend names under "ai.backend.agent" pkg namespace.
+    DOCKER = 'docker'
+    KUBERNETES = 'k8s'
+
+
 @attr.s(auto_attribs=True, slots=True)
 class VolumeInfo:
     name: str             # volume name
@@ -61,3 +67,17 @@ class ContainerLifecycleEvent:
     reason: str
     done_event: Optional[asyncio.Event] = None
     exit_code: Optional[int] = None
+    suppress_events: bool = False
+
+    def __str__(self):
+        if self.container_id:
+            cid = self.container_id[:13]
+        else:
+            cid = 'unknown'
+        return (
+            f"LifecycleEvent("
+            f"{self.event.name}, "
+            f"k:{self.kernel_id}, "
+            f"c:{cid}, "
+            f"reason:{self.reason!r})"
+        )
